@@ -5,10 +5,10 @@ import enum
 import json
 import logging
 import time
-from concurrent.futures import ThreadPoolExecutor
+from concurrent.futures import Future, ThreadPoolExecutor
 from datetime import datetime
 from itertools import chain
-from typing import Any, Dict, Iterator, List, Mapping, Optional, Sequence, Tuple, Type, TypeVar, Union
+from typing import Any, Iterable, Iterator, List, Mapping, Optional, Sequence, Type, TypeVar, Union
 from urllib.parse import urljoin
 
 import polars as pl
@@ -112,7 +112,6 @@ class _CreateOfflineQueryJobRequest(BaseModel):
         output: A list of output feature root fqns to query
         destination_format: The desired output format. Should be 'CSV' or 'PARQUET'
         input: Any givens
-        filters: Any filters
         max_samples: The maximum number of samples
     """
 
@@ -120,6 +119,8 @@ class _CreateOfflineQueryJobRequest(BaseModel):
     destination_format: str
     input: Optional[_OfflineQueryInput] = None
     max_samples: Optional[int] = None
+    observed_at_lower_bound: Optional[datetime] = None
+    observed_at_upper_bound: Optional[datetime] = None
     dataset_name: Optional[str] = None
     branch: Optional[str] = None
     environment: Optional[str] = None
@@ -132,6 +133,7 @@ class _CreateOfflineQueryJobResponse(BaseModel):
     """
 
     job_id: str
+    version: int = 1  # Deprecated
     errors: Optional[List[ChalkError]]
 
 
@@ -139,13 +141,10 @@ class _GetOfflineQueryJobResponse(BaseModel):
     """
     Attributes:
         is_finished: Whether the export job is finished (it runs asynchronously)
-        version: A version number describing the format of the output dataframes. Currently the only supported version is 1.
-            The version will be -1 if the job is not yet finished
         signed_urls: A list of signed URLs that the client can download to retrieve the exported data.
     """
 
     is_finished: bool
-    version: int
     urls: List[str]
 
 
@@ -243,9 +242,6 @@ def _expand_scalar_features_shallow(fs: Sequence[Any]) -> Iterator[str]:
 
 
 class ChalkAPIClientImpl(ChalkAPIClientProtocol):
-    _headers: Optional[Tuple[Dict[str, str], Optional[Mapping[str, str]]]]
-    _config: _ChalkClientConfig
-
     def __init__(
         self,
         *,
@@ -280,9 +276,15 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
                 active_environment=environment or token.activeEnvironment,
             )
 
-        self._headers = None
+        self._default_headers = {
+            "Accept": "application/json",
+            "Content-Type": "application/json",
+        }
+        self._engines: Optional[Mapping[str, str]] = None
+        self._exchanged_credentials = False
 
-    def _exchange_credentials(self) -> _ExchangeCredentialsResponse:
+    def _exchange_credentials(self):
+        _logger.debug("Performing a credentials exchange")
         resp = requests.post(
             url=urljoin(self._config.api_server, f"v1/oauth/token"),
             headers={
@@ -298,32 +300,19 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
         )
         response_json = resp.json()
         try:
-            return _ExchangeCredentialsResponse(**response_json)
+            creds = _ExchangeCredentialsResponse(**response_json)
         except ValidationError:
             raise HTTPError(response_json)
+        self._default_headers["Authorization"] = f"Bearer {creds.access_token}"
+        self._engines = creds.engines
+        self._exchanged_credentials = True
 
-    def _get_headers(self, environment_override: Optional[str], bust_cache: bool = False):
-        if not bust_cache and self._headers is not None:
-            _logger.debug("Returning cached authorization headers")
-            return self._headers
-
-        _logger.debug("Performing a credentials exchange")
-        creds = self._exchange_credentials()
+    def _get_headers(self, environment_override: Optional[str]):
         x_chalk_env_id = environment_override or self._config.active_environment
-        headers: Dict[str, str] = {
-            "Authorization": f"Bearer {creds.access_token}",
-            "Accept": "application/json",
-            "Content-Type": "application/json",
-        }
+        headers = dict(self._default_headers)  # shallow copy
         if x_chalk_env_id is not None:
             headers["X-Chalk-Env-Id"] = x_chalk_env_id
-
-        self._headers = (
-            headers,
-            creds.engines or {},
-        )
-
-        return self._headers
+        return headers
 
     def _request(
         self,
@@ -334,19 +323,25 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
         use_engine: bool = False,
         environment_override: Optional[str] = None,
     ) -> T:
-        x = self._get_headers(environment_override=environment_override)
-        headers, engines = x
+        # Track whether we already exchanged credentials for this request
+        exchanged_credentials = False
+        if not self._exchanged_credentials:
+            exchanged_credentials = True
+            self._exchange_credentials()
+        headers = self._get_headers(environment_override=environment_override)
         active_env = environment_override or self._config.active_environment
-        if use_engine and engines is not None and active_env in engines:
+        if use_engine and self._engines is not None and active_env in self._engines:
             assert isinstance(active_env, str)
-            base = engines[active_env]
+            base = self._engines[active_env]
         else:
             base = self._config.api_server
         url = urljoin(base, uri)
         json_body = json and json.dict()
         r = requests.request(method=method, headers=headers, url=url, json=json_body)
-        if r.status_code == 401:
-            headers, _ = self._get_headers(environment_override=environment_override, bust_cache=True)
+        if r.status_code in (401, 403) and not exchanged_credentials:
+            # It is possible that credentials expired, or that we changed permissions since we last
+            # got a token. Exchange them and try again
+            self._exchange_credentials()
             r = requests.request(method=method, headers=headers, url=url, json=json_body)
 
         r.raise_for_status()
@@ -480,17 +475,17 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
                     value=value,
                 )
 
-        input = _OfflineQueryInput(
+        query_input = _OfflineQueryInput(
             columns=columns_fqn,
             values=matrix,
         )
 
-        output = list(_expand_scalar_features_shallow(output))
+        query_output = list(_expand_scalar_features_shallow(output))
 
         try:
             response = self._create_and_await_offline_query_job(
-                output=output,
-                input=input,
+                output=query_output,
+                input=query_input,
                 dataset_name=dataset,
                 branch=branch,
                 environment=context and context.environment,
@@ -506,10 +501,10 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
         max_samples: Optional[int] = None,
         context: Optional[OfflineQueryContext] = None,
     ) -> pd.DataFrame:
-        output = _expand_scalar_features_shallow(output)
+        query_output = _expand_scalar_features_shallow(output)
 
         try:
-            response = self._create_and_await_offline_query_job(output=output, max_samples=max_samples)
+            response = self._create_and_await_offline_query_job(output=query_output, max_samples=max_samples)
         except HTTPError as e:
             _logger.debug("Got HTTP Exception while processing sample query", exc_info=e)
             raise ChalkOfflineQueryException(message=f"HTTP error while processing sample query.", errors=[])
@@ -562,12 +557,12 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
 
     def _decode_col_name(self, col_name: str) -> str:
         x_split = col_name.split("_")
-        assert len(x_split) == 2
         if x_split[0] == "ca":
+            assert len(x_split) == 2
             return x_split[1]
         elif x_split[0] == "cb":
             root_fqn_b32 = x_split[1]
-            return base64.b32decode(root_fqn_b32.replace("0", "=").upper()).decode("utf8")
+            return base64.b32decode(root_fqn_b32.replace("0", "=").upper()).decode("utf8") + "_".join(x_split[2:])
         else:
             raise ValueError(f"Unexpected identifier: {x_split[0]}")
 
@@ -583,7 +578,7 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
 
     def _create_and_await_offline_query_job(
         self,
-        output: Iterator[str],
+        output: Iterable[str],
         input: Optional[_OfflineQueryInput] = None,
         max_samples: Optional[int] = None,
         dataset_name: Optional[str] = None,
@@ -602,7 +597,7 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
         )
         response = self._create_offline_query_job(request=req)
 
-        if len(response.errors or []) > 0:
+        if response.errors is not None and len(response.errors) > 0:
             raise ChalkOfflineQueryException(message="Failed to execute offline query", errors=response.errors)
 
         while True:
@@ -613,7 +608,7 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
 
         executor = ThreadPoolExecutor(max_workers=16)
 
-        df_futures = []
+        df_futures: list[Future[pl.DataFrame]] = []
         for f in status.urls:
             df_futures.append(executor.submit(pl.read_parquet, f))
 
@@ -629,7 +624,7 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
             ts_fqn=None,
         )
         df = df.select(list(decoded_col_names.keys()))
-        df = df.rename(decoded_col_names)
+        df = df.rename(dict(decoded_col_names))
         # The parquet file is all encoded. We gotta decode
         decoded_stmts = []
         for col in df.columns:
@@ -640,7 +635,7 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
                 def _bound_decode(val: Any, col: str = col):
                     if val is None:
                         return None
-                    return FEATURE_CODEC.decode_fqn(col, json.loads(val))
+                    return json.loads(val)
 
                 decoded_stmts.append(
                     pl.col(col).apply(_bound_decode, return_dtype=FEATURE_CODEC.get_polars_dtype(fqn=col))
