@@ -1,12 +1,17 @@
 from __future__ import annotations
 
+import inspect
+
 import collections.abc
+import dataclasses
 import datetime
 import enum
 import functools
+import inspect
 import operator
 import pathlib
 import warnings
+from pydantic import BaseModel
 from typing import (
     TYPE_CHECKING,
     Any,
@@ -27,12 +32,14 @@ from typing import (
 )
 
 import polars as pl
+from pydantic import BaseModel
 
 try:
     import zoneinfo
 except ImportError:
     # Zoneinfo was introduced in python 3.9
     from backports import zoneinfo
+
 from dateutil import parser
 
 from chalk.features.feature_field import Feature
@@ -60,10 +67,6 @@ else:
 TDataFrame = TypeVar("TDataFrame", bound="DataFrame")
 
 
-def _polars_data_frames_equal(a: pl.DataFrame, b: pl.DataFrame):
-    return a.frame_equal(b)
-
-
 def _iso_parse(x: str, expected_tz: datetime.tzinfo | None) -> datetime.datetime:
     """This function converts the input into an expected timezone, and drops it, so we can cast it in polars with with_time_zone"""
     parsed = parser.isoparse(x)
@@ -102,6 +105,7 @@ class DataFrameMeta(type):
         new_filters: List[Filter] = []
         new_references_feature_set: Optional[Type[Features]] = None
         new_columns: List[Feature] = []
+        pydantic_model = None
 
         for a in item:
             if isinstance(a, Filter):
@@ -122,6 +126,8 @@ class DataFrameMeta(type):
                 # into expressions that return filters instead of booleans
                 # This function will be called again for this annotation, so we can ignore it for now.
                 pass
+            elif inspect.isclass(a) and issubclass(a, BaseModel):
+                pydantic_model = a
             else:
                 raise TypeError(f"Invalid type for DataFrame[{a}]: {type(a)}")
 
@@ -139,6 +145,7 @@ class DataFrameMeta(type):
             filters = tuple(new_filters)
             __columns__ = tuple(new_columns)
             __references_feature_set__ = new_references_feature_set
+            __pydantic_model__ = pydantic_model
 
             def __new__(cls: Type[TDataFrame], *args: Any, **kwargs: Any) -> TDataFrame:
                 raise RuntimeError(
@@ -204,6 +211,7 @@ class DataFrame(metaclass=DataFrameMeta):
     __columns__: ClassVar[Tuple[Feature, ...]] = ()
     references_feature_set: Optional[Type[Features]]  # set via a @property on the metaclass
     __references_feature_set__: ClassVar[Optional[Type[Features]]] = None
+    __pydantic_model__: ClassVar[Optional[Type[BaseModel]]] = None
 
     def __init__(
         self,
@@ -217,6 +225,7 @@ class DataFrame(metaclass=DataFrameMeta):
         ] = None,
         pandas_dataframe: Optional[pd.DataFrame] = None,  # For backwards compatibility
         convert_dtypes: bool = True,  # By default, data should match the dtype of the feature. However, when doing comparisions, data will be converted to bools, in which case it should no longer be validated.
+        pydantic_model: Optional[Type[BaseModel]] = None,
     ):
         """Construct a Chalk DataFrame
 
@@ -228,6 +237,8 @@ class DataFrame(metaclass=DataFrameMeta):
         """
         from chalk.features.feature_set import Features, FeatureSetBase
 
+        self._pydantic_model = pydantic_model
+
         # Typing the keys of ``data`` as Any, as {FeatureCls.item: x} would be typed as the underlying annotation of the features cls
         if pandas_dataframe is not None:
             warnings.warn(
@@ -345,14 +356,16 @@ class DataFrame(metaclass=DataFrameMeta):
             raise ValueError(f"Unable to convert data of type {type(data).__name__} into a DataFrame")
         # Rename / validate that all column names are root fqns
         # It is possible that a feature name is a
-        self.columns = tuple(Feature.from_root_fqn(str(c)) for c in underlying.columns)
-        underlying = underlying.rename(
-            {original_c: new_c.root_fqn for (original_c, new_c) in zip(underlying.columns, self.columns)}
-        )
-        namespaces = [x.path[0].parent.namespace if len(x.path) > 0 else x.namespace for x in self.columns]
+        if self._pydantic_model is None:
+            self.columns = tuple(Feature.from_root_fqn(str(c)) for c in underlying.columns)
+            underlying = underlying.rename(
+                {original_c: new_c.root_fqn for (original_c, new_c) in zip(underlying.columns, self.columns)}
+            )
+        else:
+            self.columns = ()
 
         # Convert columns to the correct dtype to match the fqn
-        if convert_dtypes:
+        if self._pydantic_model is None and convert_dtypes:
             for root_fqn in underlying.columns:
                 expected_dtype = FEATURE_CODEC.get_polars_dtype(root_fqn)
                 actual_dtype = underlying.schema[root_fqn]
@@ -398,14 +411,15 @@ class DataFrame(metaclass=DataFrameMeta):
         self._underlying: pl.LazyFrame = underlying
 
         # Remove the pseudo-features when determining the namespace
+        namespaces = (x.path[0].parent.namespace if len(x.path) > 0 else x.namespace for x in self.columns)
         namespaces_set = set(x for x in namespaces if not x.startswith("__chalk__"))
-        if len(namespaces_set) != 1:
+        if len(namespaces_set) == 1:
+            self.namespace = get_unique_item(namespaces_set, f"dataframe column namespaces")
+            self.references_feature_set = FeatureSetBase.registry[self.namespace]
+        else:
             # Allow empty dataframes or dataframes with multiple namespaces
             self.namespace = None
             self.references_feature_set = None
-        else:
-            self.namespace = get_unique_item(namespaces_set, f"dataframe column namespaces")
-            self.references_feature_set = FeatureSetBase.registry[self.namespace]
 
     ##############
     # Classmethods
@@ -486,26 +500,26 @@ class DataFrame(metaclass=DataFrameMeta):
     #############
 
     def max(self):
-        return DataFrame(self._underlying.max(), convert_dtypes=False)
+        return DataFrame(self._underlying.max(), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
     def mean(self):
-        return DataFrame(self._underlying.mean(), convert_dtypes=False)
+        return DataFrame(self._underlying.mean(), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
     def median(self):
-        return DataFrame(self._underlying.median(), convert_dtypes=False)
+        return DataFrame(self._underlying.median(), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
     def min(self):
-        return DataFrame(self._underlying.min(), convert_dtypes=False)
+        return DataFrame(self._underlying.min(), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
     def std(self, ddof: int = 1):
-        return DataFrame(self._underlying.std(ddof), convert_dtypes=False)
+        return DataFrame(self._underlying.std(ddof), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
     def sum(self):
         # Treat missing sums as zero
-        return DataFrame(self._underlying.sum().fill_null(0), convert_dtypes=False)
+        return DataFrame(self._underlying.sum().fill_null(0), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
     def var(self, ddof: int = 1):
-        return DataFrame(self._underlying.var(ddof), convert_dtypes=False)
+        return DataFrame(self._underlying.var(ddof), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
     ####################
     # Summary Operations
@@ -589,7 +603,11 @@ class DataFrame(metaclass=DataFrameMeta):
             other = other.collect()
         if isinstance(other, pd.DataFrame):
             other = pl.from_pandas(other)
-        return DataFrame(op(materialized, other), convert_dtypes=convert_dtypes)
+        return DataFrame(
+            op(materialized, other),
+            convert_dtypes=convert_dtypes,
+            pydantic_model=self._pydantic_model,
+        )
 
     def __eq__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]):  # type: ignore
         return self._perform_op(operator.eq, other, convert_dtypes=False)
@@ -692,13 +710,13 @@ class DataFrame(metaclass=DataFrameMeta):
             # Do the projection
             if len(projections) > 0:
                 df = df.select(projections)
-            return DataFrame(df)
+            return DataFrame(df, pydantic_model=self)
         else:
             # Otherwise, use the standard polars selection format
             # Must materialize the dataframe to use __getitem__
             materialized = self._materialize()
             df = materialized[item]
-            return DataFrame(df)
+            return DataFrame(df, pydantic_model=self._pydantic_model)
 
 
 def _feature_type_or_value(e: Union[Feature, FeatureWrapper]):
