@@ -17,6 +17,7 @@ from pendulum.tz.timezone import Timezone
 from polars.internals.construction import is_namedtuple
 
 from chalk.serialization.parsed_annotation import ParsedAnnotation
+from chalk.streams import Windowed
 from chalk.utils.enum import get_enum_value_type
 
 if TYPE_CHECKING:
@@ -210,9 +211,11 @@ class FeatureCodec:
             dtype = pl.List(dtype)
         return dtype
 
-    def _get_pl_dtype_for_type(self, underlying: type, fqn: str) -> Union[pl.DataType, Type[pl.DataType]]:
-        if not isinstance(underlying, type):
-            raise
+    def _get_pl_dtype_for_type(self, underlying: type | Windowed, fqn: str) -> Union[pl.DataType, Type[pl.DataType]]:
+        if isinstance(underlying, Windowed):
+            # If the parent window feature appears in a dataframe, then it's likely in a streaming resolver
+            # Use the underlying type for that windowed feature
+            return self._get_pl_dtype_for_type(underlying._kind, fqn)
         if issubclass(underlying, enum.Enum):
             # For enums, require all members to have the same type
             underlying = get_enum_value_type(underlying)
