@@ -1,6 +1,9 @@
-from typing import Any, Sequence, Set, Type, Union
+from typing import Any, List, Sequence, Set, Type, Union
 
+import polars
+import pyarrow
 from pydantic import BaseModel
+from typing_extensions import TypeAlias
 
 from chalk.utils import AnyDataclass
 
@@ -13,8 +16,19 @@ class StreamResolverParamMessage(StreamResolverParam):
     typ: Union[Type[str], Type[bytes], Type[BaseModel], AnyDataclass]
 
 
+StreamResolverWindowType: TypeAlias = Union[
+    Type[List[str]],
+    Type[List[bytes]],
+    Type[List[BaseModel]],
+    Type[AnyDataclass],
+    Type[pyarrow.Table],
+    Type[polars.DataFrame],
+    Any,  # The annotation value is likely going to be a GenericAlias, which messes with pydantic's validation
+]
+
+
 class StreamResolverParamMessageWindow(StreamResolverParam):
-    item_typ: Union[Type[str], Type[bytes], Type[BaseModel], AnyDataclass]
+    typ: StreamResolverWindowType
 
 
 class StreamResolverSignature(BaseModel):
